@@ -13,7 +13,7 @@ use crate::ir::Inst as IRInst;
 use crate::ir::{Opcode, Type};
 use crate::machinst::lower::*;
 use crate::machinst::*;
-use crate::CodegenResult;
+use crate::{CodegenError, CodegenResult};
 
 use crate::isa::aarch64::inst::*;
 use crate::isa::aarch64::AArch64Backend;
@@ -1098,55 +1098,98 @@ pub(crate) fn lower_vector_compare<C: LowerCtx<I = Inst>>(
     ty: Type,
     cond: Cond,
 ) -> CodegenResult<()> {
-    let is_float = match ty {
-        F32X4 | F64X2 => true,
-        _ => false,
-    };
+    let is_float = ty.lane_type().is_float();
     let size = VectorSize::from_ty(ty);
-    // 'Less than' operations are implemented by swapping
-    // the order of operands and using the 'greater than'
-    // instructions.
-    // 'Not equal' is implemented with 'equal' and inverting
-    // the result.
-    let (alu_op, swap) = match (is_float, cond) {
-        (false, Cond::Eq) => (VecALUOp::Cmeq, false),
-        (false, Cond::Ne) => (VecALUOp::Cmeq, false),
-        (false, Cond::Ge) => (VecALUOp::Cmge, false),
-        (false, Cond::Gt) => (VecALUOp::Cmgt, false),
-        (false, Cond::Le) => (VecALUOp::Cmge, true),
-        (false, Cond::Lt) => (VecALUOp::Cmgt, true),
-        (false, Cond::Hs) => (VecALUOp::Cmhs, false),
-        (false, Cond::Hi) => (VecALUOp::Cmhi, false),
-        (false, Cond::Ls) => (VecALUOp::Cmhs, true),
-        (false, Cond::Lo) => (VecALUOp::Cmhi, true),
-        (true, Cond::Eq) => (VecALUOp::Fcmeq, false),
-        (true, Cond::Ne) => (VecALUOp::Fcmeq, false),
-        (true, Cond::Mi) => (VecALUOp::Fcmgt, true),
-        (true, Cond::Ls) => (VecALUOp::Fcmge, true),
-        (true, Cond::Ge) => (VecALUOp::Fcmge, false),
-        (true, Cond::Gt) => (VecALUOp::Fcmgt, false),
-        _ => unreachable!(),
-    };
 
-    if swap {
-        std::mem::swap(&mut rn, &mut rm);
-    }
+    if is_float && (cond == Cond::Vc || cond == Cond::Vs) {
+        let tmp = ctx.alloc_tmp(ty).only_reg().unwrap();
 
-    ctx.emit(Inst::VecRRR {
-        alu_op,
-        rd,
-        rn,
-        rm,
-        size,
-    });
-
-    if cond == Cond::Ne {
-        ctx.emit(Inst::VecMisc {
-            op: VecMisc2::Not,
+        ctx.emit(Inst::VecRRR {
+            alu_op: VecALUOp::Fcmeq,
             rd,
-            rn: rd.to_reg(),
+            rn,
+            rm: rn,
             size,
         });
+        ctx.emit(Inst::VecRRR {
+            alu_op: VecALUOp::Fcmeq,
+            rd: tmp,
+            rn: rm,
+            rm,
+            size,
+        });
+        ctx.emit(Inst::VecRRR {
+            alu_op: VecALUOp::And,
+            rd,
+            rn: rd.to_reg(),
+            rm: tmp.to_reg(),
+            size,
+        });
+
+        if cond == Cond::Vs {
+            ctx.emit(Inst::VecMisc {
+                op: VecMisc2::Not,
+                rd,
+                rn: rd.to_reg(),
+                size,
+            });
+        }
+    } else {
+        // 'Less than' operations are implemented by swapping
+        // the order of operands and using the 'greater than'
+        // instructions.
+        // 'Not equal' is implemented with 'equal' and inverting
+        // the result.
+        let (alu_op, swap) = match (is_float, cond) {
+            (false, Cond::Eq) => (VecALUOp::Cmeq, false),
+            (false, Cond::Ne) => (VecALUOp::Cmeq, false),
+            (false, Cond::Ge) => (VecALUOp::Cmge, false),
+            (false, Cond::Gt) => (VecALUOp::Cmgt, false),
+            (false, Cond::Le) => (VecALUOp::Cmge, true),
+            (false, Cond::Lt) => (VecALUOp::Cmgt, true),
+            (false, Cond::Hs) => (VecALUOp::Cmhs, false),
+            (false, Cond::Hi) => (VecALUOp::Cmhi, false),
+            (false, Cond::Ls) => (VecALUOp::Cmhs, true),
+            (false, Cond::Lo) => (VecALUOp::Cmhi, true),
+            (true, Cond::Eq) => (VecALUOp::Fcmeq, false),
+            (true, Cond::Ne) => (VecALUOp::Fcmeq, false),
+            (true, Cond::Mi) => (VecALUOp::Fcmgt, true),
+            (true, Cond::Ls) => (VecALUOp::Fcmge, true),
+            (true, Cond::Ge) => (VecALUOp::Fcmge, false),
+            (true, Cond::Gt) => (VecALUOp::Fcmgt, false),
+            _ => {
+                return Err(CodegenError::Unsupported(format!(
+                    "Unsupported {} SIMD vector comparison: {:?}",
+                    if is_float {
+                        "floating-point"
+                    } else {
+                        "integer"
+                    },
+                    cond
+                )))
+            }
+        };
+
+        if swap {
+            std::mem::swap(&mut rn, &mut rm);
+        }
+
+        ctx.emit(Inst::VecRRR {
+            alu_op,
+            rd,
+            rn,
+            rm,
+            size,
+        });
+
+        if cond == Cond::Ne {
+            ctx.emit(Inst::VecMisc {
+                op: VecMisc2::Not,
+                rd,
+                rn: rd.to_reg(),
+                size,
+            });
+        }
     }
 
     Ok(())
@@ -1364,7 +1407,7 @@ pub(crate) fn lower_i64x2_mul<C: LowerCtx<I = Inst>>(c: &mut C, insn: IRInst) {
     //   rd = |dg+ch|be+af||dg+ch|be+af|
     c.emit(Inst::VecRRR {
         alu_op: VecALUOp::Addp,
-        rd: rd,
+        rd,
         rn: rd.to_reg(),
         rm: rd.to_reg(),
         size: VectorSize::Size32x4,
@@ -1740,6 +1783,22 @@ pub(crate) fn is_valid_atomic_transaction_ty(ty: Type) -> bool {
     }
 }
 
+pub(crate) fn emit_atomic_load<C: LowerCtx<I = Inst>>(
+    ctx: &mut C,
+    rt: Writable<Reg>,
+    insn: IRInst,
+) {
+    assert!(ctx.data(insn).opcode() == Opcode::AtomicLoad);
+    let inputs = insn_inputs(ctx, insn);
+    let rn = put_input_in_reg(ctx, inputs[0], NarrowValueMode::None);
+    let access_ty = ctx.output_ty(insn, 0);
+    assert!(is_valid_atomic_transaction_ty(access_ty));
+    // We're ignoring the result type of the load because the LoadAcquire will
+    // explicitly zero extend to the nearest word, and also zero the high half
+    // of an X register.
+    ctx.emit(Inst::LoadAcquire { access_ty, rt, rn });
+}
+
 fn load_op_to_ty(op: Opcode) -> Option<Type> {
     match op {
         Opcode::Sload8 | Opcode::Uload8 | Opcode::Sload8Complex | Opcode::Uload8Complex => Some(I8),
@@ -1769,14 +1828,14 @@ fn load_op_to_ty(op: Opcode) -> Option<Type> {
 /// a load can sometimes be merged into another operation.
 pub(crate) fn lower_load<
     C: LowerCtx<I = Inst>,
-    F: FnMut(&mut C, ValueRegs<Writable<Reg>>, Type, AMode),
+    F: FnMut(&mut C, ValueRegs<Writable<Reg>>, Type, AMode) -> CodegenResult<()>,
 >(
     ctx: &mut C,
     ir_inst: IRInst,
     inputs: &[InsnInput],
     output: InsnOutput,
     mut f: F,
-) {
+) -> CodegenResult<()> {
     let op = ctx.data(ir_inst).opcode();
 
     let elem_ty = load_op_to_ty(op).unwrap_or_else(|| ctx.output_ty(ir_inst, 0));
@@ -1785,7 +1844,7 @@ pub(crate) fn lower_load<
     let mem = lower_address(ctx, elem_ty, &inputs[..], off);
     let rd = get_output_reg(ctx, output);
 
-    f(ctx, rd, elem_ty, mem);
+    f(ctx, rd, elem_ty, mem)
 }
 
 pub(crate) fn emit_shl_i128<C: LowerCtx<I = Inst>>(
